
from django.conf import settings

from .component import (  # noqa
    Component, AuthComponent, StaffComponent, on_commit, broadcast
)


if getattr(settings, 'REACTOR_AUTO_BROADCAST', False):
    import reactor.auto_broadcast  # noqa



import json
from uuid import uuid4
from contextlib import asynccontextmanager

from pyquery import PyQuery as q

from django.contrib.auth.models import AnonymousUser
from django.core.serializers.json import DjangoJSONEncoder
from channels.testing import WebsocketCommunicator

from .channels import ReactorConsumer


class ReactorCommunicator(WebsocketCommunicator):
    MAX_WAIT = 2  # seconds

    def __init__(self, user=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scope['user'] = user or AnonymousUser()
        self._components = {}
        self.redirected_to = None
        self.loop_timeout = None

    async def auto_join(self, response):
        doc = q(response.content)
        for component in doc('[id][state]'):
            tag_name = component.get('is') or component.tag
            state = json.loads(component.get('state'))
            component_id = self.add_component(tag_name, state)
            await self.send_join(component_id)

    def add_component(self, *args, **kwargs):
        component = Component(*args, **kwargs)
        self._components[component.id] = component
        return component.id

    def __getitem__(self, _id):
        return self._components[_id]

    def get_by_name(self, name):
        for component in self._components.values():
            if component.tag_name == name:
                return component

    async def send_join(self, component_id):
        component = self._components[component_id]
        await self.send_json_to({
            'command': 'join',
            'payload': {
                'tag_name': component.tag_name,
                'state': component.state,
            },
        })
        await self.loop_over_messages()
        return component.doc

    async def send(self, _id, _name, **state):
        assert _id in self._components
        await self.send_json_to({
            'command': 'user_event',
            'payload': {
                'name': _name,
                'state': dict(state, id=_id)
            }
        })
        await self.loop_over_messages(reset_timeout=True)
        return self._components[_id].doc

    async def send_json_to(self, data):
        await self.send_to(text_data=json.dumps(data, cls=DjangoJSONEncoder))

    async def loop_over_messages(self, reset_timeout=False):
        if reset_timeout or not self.loop_timeout:
            self.loop_timeout = 0.1
            while await self.receive_nothing(timeout=self.loop_timeout):
                self.loop_timeout *= 2
                if self.loop_timeout > self.MAX_WAIT:
                    break

        while not await self.receive_nothing(timeout=self.loop_timeout):
            response = await self.receive_json_from()
            if response['type'] in ('redirect', 'push_state'):
                self.redirected_to = response['url']
            else:
                component = self._components[response['id']]
                if response['type'] == 'render':
                    component.apply_diff(response['html_diff'])
                elif response['type'] == 'remove':
                    component.apply_remove()


@asynccontextmanager
async def reactor(consumer=ReactorConsumer, path='/reactor', user=None):
    comm = ReactorCommunicator(application=consumer, path=path, user=user)
    connected, subprotocol = await comm.connect()
    assert connected
    try:
        yield comm
    finally:
        await comm.disconnect()


class Component:
    def __init__(self, tag_name: str, state: dict = None):
        state = state or {}
        state.setdefault('id', str(uuid4()))
        self.tag_name = tag_name
        self.state = state
        self.last_received_html = ''
        self.doc = None
        self.removed = False

    @property
    def id(self):
        return self.state['id']

    def apply_diff(self, html_diff):
        html = []
        cursor = 0
        for diff in html_diff:
            if isinstance(diff, str):
                html.append(diff)
            elif diff < 0:
                cursor -= diff
            else:
                html.append(self.last_received_html[cursor:cursor + diff])
                cursor += diff
        self.last_received_html = ''.join(html)
        self.doc = q(self.last_received_html)

        state = self.doc.attr['state']
        if state:
            self.state = json.loads(state)

    def apply_remove(self):
        self.removed = True

    def __str__(self):
        return f'<{self.tag_name} {self.state}>'

    __repr__ = __str__



from os.path import dirname, abspath

from setuptools import find_packages, setup


HERE = abspath(dirname(__file__))
README = open('README.md', encoding='utf-8').read()

setup(
    name='django-reactor',
    version='1.3.4b0',
    url='https://github.com/edelvalle/reactor',
    author='Eddy Ernesto del Valle Pino',
    author_email='eddy@edelvalle.me',
    long_description=README,
    long_description_content_type='text/markdown',
    description="Brings LiveView from Phoenix framework into Django",
    license='BSD',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.6',
    install_requires=[
        'channels>=2.2.0,<2.5',
        'diff-match-patch>=20181111',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import json
from django import template
from django.db.models import QuerySet, Model
from django.template.loader import render_to_string
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.safestring import mark_safe


from ..component import Component

register = template.Library()


class ReactorJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        if isinstance(o, Model):
            return o.id
        if isinstance(o, QuerySet):
            return list(o.values_list('id', flat=True))
        return super().default(o)


@register.simple_tag
def reactor_header():
    return render_to_string(
        'reactor_header.html',
        {
            'components': [
                (name, c.extends) for name, c in Component._all.items()
            ]
        },
    )


@register.simple_tag(takes_context=True)
def component(context, _name, id=None, **kwargs):
    parent = context.get('this')  # type: Component
    if parent:
        component = parent._children.get_or_create(_name, id=id, **kwargs)
    else:
        component = Component.build(_name, context=context, id=id)
        component.mount(**kwargs)
    return component.render()


@register.filter
def concat(value, arg):
    return f'{value}-{arg}'


@register.filter()
def tojson_safe(value):
    return mark_safe(tojson(value))


@register.filter()
def tojson(value):
    return json.dumps(value, cls=ReactorJSONEncoder)


@register.filter
def then(value, true_result):
    if value:
        return true_result
    else:
        return ''


@register.filter
def ifnot(value, false_result):
    if not value:
        return false_result
    else:
        return ''
